/* Homework Assignment #1 EECS 489 F05 */
/* Ilya Wagner (ivagner@eecs.umich.edu) */

/*
 * Copyright (c) 1997, 1998, 2000 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
 *         Eric Cronin (ecronin@eecs.umich.edu)
 *         Wenjie Wang (wenjiew@eecs.umich.edu)
 *         Matt England (moenglan@eecs.umich.edu) 
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifdef _TURNIN_SUN
#include <crypt.h>
#endif
#include <unistd.h>

#include "turnin.h"
#include "crypto.h"

extern int errno;

/* stubs for the functions provided in _turnin.o */
extern int  _tin_setup_conn(char *, int);
extern void _tin_send_hdr(int, char *, short, char *, int);
extern int  _tin_send_fhdr(int, char *, int *);
extern void _tin_send_file(int, int);
extern int _tin_recv_ack(int);

void
tin_usage(char *progname)
{
  fprintf(stderr, 
    "Usage: %s [-r] [-s servername] [-p portnum] [-l] -a assignment [files...]\n", 
    progname);
  exit(errno);
}
   
/*
 * tin_args: Parse args. (command line)
 *
 * Return 0 on success or errno on failure.
 * On successful return, "servname" contains the
 * server's host name, "port" contains the
 * port# to use, in host byte order, "asgn"
 * contains the assignment name, and "files"
 * contains a list of filenames.
*/
int
tin_args(int argc, char *argv[], char **servname, int *port,
	 short *op, char **asgn,
	 short *nfiles, char ***files)
{
  extern char *optarg;
  extern int optind;
  char c;

  errno = 0;

  *asgn = NULL;
  *servname = TIN_SRVR;
  *port = TIN_PORT;
  *op = TIN_PUT;
 
  while ((c = getopt(argc, argv, "s:p:rla:")) != EOF) {
    switch (c) {
    case 's':
      *servname = optarg;
      break;
    case 'p':
      *port = atoi(optarg);
      break;
    case 'r':
      *nfiles = 0;
      *op = TIN_SETPASS;
      break;
    case 'l':
      *nfiles = 0;
      *op = TIN_LST;
      break;
    case 'a':
      *asgn = optarg;
      break;
    default:
      errno = EINVAL;
      break;
    }
  }
  
  if (errno != 0) {
    return(errno);
  }

  /* Finish up w/ assignment # and files */
  if (!*asgn) {
    if (*op != TIN_SETPASS) errno = EINVAL;
    else {
      *asgn = (char *)malloc(TIN_ALEN);
      memset(*asgn,0,TIN_ALEN);
      memset(*asgn,'a',TIN_ALEN - 1);
    }
  } else if ((argc - optind) >= 1) {
    /* if files is specified */
    *nfiles = argc - optind;
    *files = &argv[optind];
    *op = TIN_PUT;
  } else if (*op != TIN_LST) {
    errno = EINVAL;
  }

  return(errno);
}

/*
 * tin_setup_conn: setup TCP connection to server on
 * host "servname" at port "port".  The port "port"
 * is given in host byte-order.
 *
 * On system error, terminate process.
 * On sucess, return socket id.  Socket is open
 * for synchronous communication.
*/
int
tin_setup_conn(char *servname, int port)
{     
//  return _tin_setup_conn(servname, port); 

  struct sockaddr_in server_addr;
  int sock, sockopt_on=1;
  struct hostent * he;
  
  //get a socket
  if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) 
  {
    perror("socket acquisition failed");
    exit(1);
  }


  // make socket reusable
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, & sockopt_on, sizeof (int)) == -1)
  {
  perror ("socket reusable option failed");
  exit (1);
  }
  
  // look up host
  he = gethostbyname (servname);
  if (he == NULL)
  	{
	perror ("name lookup failed ");
	exit (1);
	}
  
  // prepare server address datastructure
  memset (( char *) &server_addr, 0, sizeof (struct sockaddr_in));
  server_addr.sin_family = PF_INET;
  server_addr.sin_port = htons (port);
  server_addr.sin_addr = *((struct in_addr *)he->h_addr);


  // connect to server
  if (connect (sock, (struct sockaddr *) &server_addr, sizeof (struct sockaddr_in)) == -1)
  	{
	perror ("connection failed");
	exit(1);
	} 
	
  return sock;	
}
  
/*
 * tin_send_hdr: send turnin header to server on socket "sd".
 * Arg "progname" contains the program's name (for error reporting
 * purposes), "asgn" the assignment's name, "op" one of TIN_PUT or TIN_LST,
 * and "nfiles" the number of files to send.  Check that "asgn"
 * is exactly TIN_ALEN-1 long.  Grab user's uniqname from the
 * environment using getenv(3C).
 *
 * On error, terminate process.
*/
void
tin_send_hdr(int sd, char *progname, short op, char *asgn, 
	     int nfiles)
{
 //return _tin_send_hdr(sd, progname, op, asgn, nfiles);

 struct _tin_head header;
 int bytes_to_send= sizeof (struct _tin_head);
 int bytes_sent=0;
   
 //make sure asgn has TIN_ALEN-1 chars. 
  if (strlen(asgn) != (TIN_ALEN-1))
 	{
	perror ("assignment name length");
	exit(1);
	}
  

 header.tih_vers = htons(TIN_VERS);
 header.tih_op = htons(op);
 memcpy(&header.tih_asgn,asgn,TIN_ALEN-1);

 header.tih_asgn[TIN_ALEN-1]='\0';
  
 // 0 for TIN_LIST taken care in parse
 header.tih_nfiles=htons((unsigned short)nfiles);
 
 // now need to send the header
 while (bytes_sent < bytes_to_send)
 	{
	int sent=send (sd, ((char *)(& header)+bytes_sent), bytes_to_send-bytes_sent,0);
		if (sent==-1)
			{
			perror ("tin_send_hdr");
			exit(1);
			}
	bytes_sent+=sent;
	}
 
} 

     
/*
 * tin_send_fhdr: send file header to server on socket "sd".
 * Arg "fname" contains the file name.
 *
 * On sucess, the file named is opened for reading and
 * its descriptor returned.  The arg "fsize" contains
 * the size of the file.  
 *
 * On error, terminate process.
*/ 
int
tin_send_fhdr(int sd, char *fname, int *fsize)
{
//  return _tin_send_fhdr(sd, fname, fsize); 

 struct _tif_head header;
 int bytes_to_send=sizeof (struct _tif_head);
 int bytes_sent=0;
 int file;
 
 // open the file for reading
 file=open(fname,O_RDONLY);
 if (file==-1)
 		{
		perror ("cannot open file");
		exit(1);
		}

 // get file size and rewind
 *fsize=(int)lseek (file, 0, SEEK_END);
 lseek (file, 0, SEEK_SET);


 // prepare packet
 memset(&header,0,bytes_to_send);
 memcpy(&header.tif_fname,fname,strlen(fname));
 header.tif_fsize=htonl(*fsize);

 // send header
 while (bytes_sent < bytes_to_send)
 	{
	int sent=send (sd, ((char *)(&header)+bytes_sent), bytes_to_send-bytes_sent,0);
		if (sent==-1)
			{
			perror ("tin_send_fhdr");
			exit(1);
			}
	bytes_sent+=sent;
	} 

 return file;
}

/*
 * tin_send_file: read data from file fd and
 * send it to server connected on socket sd.
 * Send data on chunks of TIN_DLEN at a time.
 *
 * On error, terminate process.
*/ 
void
tin_send_file(int sd, int fd)
{
 
// return _tin_send_file(sd, fd); 

 // get file size again
 int total_to_send = (int)lseek (fd, 0, SEEK_END);
 int bytes_to_send;
 int bytes_sent=0;
 char buff [TIN_DLEN];

 lseek (fd, 0, SEEK_SET); // rewind

  while(read(fd, buff, TIN_DLEN)!=0)
  {
   bytes_sent=0;
   bytes_to_send = total_to_send < TIN_DLEN ? total_to_send : TIN_DLEN;
   
   // send chunk
   while (bytes_sent < bytes_to_send)
 	{
	int sent=send (sd, ((char *)(& buff)+bytes_sent), bytes_to_send-bytes_sent,0);
		if (sent==-1)
			{
			perror ("tin_send_file");
			exit(1);
			}
	bytes_sent+=sent;
	}    

  total_to_send -= bytes_sent;

  }
  

}
  
/*
 * tin_recv_ack: synchronously receive acknowledgement
 * from server for files sent.  Message from server 
 * is printed out on stderr.  
 *
 * Return number of bytes received.
*/ 
int
tin_recv_ack(int sd)
{

  //return(_tin_recv_ack(sd));
    
  int bytes_received=0;
  char buff_curr, buff_prev=0;
  int recvd;

  do {
	recvd=recv (sd, & (buff_curr), 1,0);
		if (recvd==-1)
			{
			perror ("tin_recv_ack");
			exit(1);
			}
	bytes_received +=recvd;
	
	fprintf(stderr,"%c",buff_curr);		
	
	buff_prev=buff_curr;
	} while (recvd!=0);


  return bytes_received;
}

/*
 * Sends the new password.
 */
int tin_send_setpass(int sd, char * newpass, int len) {
  int err;
  err = send(sd,newpass,len,0);
  tin_check_err(err < 0, "sending new password");
  return err - len;
}

/**
 * tin_handshake: Perform the client-side handshake.
 * 
 * Return 0 on success, errno on failure.
 */
int tin_handshake(int sd, char * password) {
  char buf[TIND_CHALLENGE_LEN];
  char * token;
  char uniq[9];
  char nonce[3];
  int tosend, sent, torecv, recved, ptr;
  
  strcpy(uniq, getenv("USER"));
  memset(buf,0,TIND_CHALLENGE_LEN);
  memset(nonce,0,3);
  if (create_nonce(nonce)) {
    return errno;
  }
  sprintf(buf,"%s",uniq);
  strncpy(&(buf[TIN_CHALLENGE_LEN-2]),nonce,2);

  tosend = TIN_CHALLENGE_LEN;
  sent = 0;
  while (tosend && !(sent < 0)) {
    sent = send(sd, &(buf[sent]), tosend, 0);
    tin_check_err(sent < 0, "send");
    tosend -= sent;
  }
  if (sent < 0) {
    return(errno);
  }
  
  torecv = TIND_CHALLENGE_LEN;
  recved = 0; ptr=0;
  while (torecv && !(recved < 0)) {
    recved = recv(sd, &(buf[ptr]),torecv,0);
    tin_check_err(sent < 0,"send");
    torecv -= recved;
    ptr+=recved;
  }
  if (recved < 0) {
    return errno;
  }
  
  //check the response
  token = crypt(password,nonce);
  if (strncmp(token,buf,CRYPT_LEN)) {
    return(errno = EINVAL);
  }
  strncpy(nonce,&(buf[TIND_CHALLENGE_LEN - 2]),2);
  nonce[2] = '\0'; //crypt expects a string
  token = crypt(password,nonce);
  
  tosend = CRYPT_LEN;
  sent = 0;
  while (tosend && !(sent < 0)) {
    sent = send(sd, token + sent, tosend, 0);
    tin_check_err(sent < 0, "send");
    tosend -= sent;
  }
  if (sent < 0) {
    return(errno);
  }
  return 0;
}

int
main(int argc, char *argv[])
{
  char **files;
  int i, sd;
  int fd, fsize;
  int port;
  short op, nfiles;
  char *servname, *asgn;
  char * password;
  char * newpass;
  
  if (tin_args(argc, argv, &servname, &port, 
      &op, &asgn, &nfiles, &files)) {
    tin_usage(argv[0]);
  }
  
  if (op == TIN_SETPASS) {
    char * temp;
    temp = getpassword("Current password:");
    password = malloc(strlen(temp)+1);
    memset(password,0,strlen(temp)+1);
    strcpy(password,temp);
    newpass = getpassword("New password:");
    nfiles = strlen(newpass);
    encrypt_pass(newpass,password,nfiles);
  } else {
    password = getpassword("Password:");
  }
 
  // get socket
  sd = tin_setup_conn(servname, port);

  if (tin_handshake(sd,password)) {
    close(sd);
    fprintf(stderr,"Authentication failed. Did you type your password correctly?\n");
    fprintf(stderr,"If the problem persists, contact the instructor.\n");
    exit(1);
  }

  tin_send_hdr(sd, argv[0], op, asgn, nfiles); //must check OP?
  
  if (op == TIN_SETPASS) {
    if (tin_send_setpass(sd,newpass,nfiles)) {
      fprintf(stderr,"Failed to reset password.\n");
    }
  } else {
   for (i = 0; i < nfiles; i++) 
      {
      fd = tin_send_fhdr(sd, files[i], &fsize);
      tin_send_file(sd, fd);
      }
  }

  if (!tin_recv_ack(sd)) {
    fprintf(stderr, "%s: Alert! Server did not acknowledge receipt.\n", argv[0]);
  }

  close(sd);
  exit(0);
}
