/* Homework Assignment #1 EECS 489 F05 */
/* Ilya Wagner (ivagner@eecs.umich.edu) */

/*
 * Copyright (c) 1997-2001 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin (jamin@eecs.umich.edu)
 *         Eric Cronin (ecronin@eecs.umich.edu)
 *         Wenjie Wang (wenjiew@eecs.umich.edu)
 *         Matt England (moenglan@eecs.umich.edu) 
 *
*/
#include <sys/select.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <values.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef _TURNIN_SUN
#include <crypt.h>
#endif

#include "turnin.h"
#include "crypto.h"

#define TIND_QLEN 200
#define TIND_ERRSLEEPSEC 2
#define TIND_FLIST TIN_FLEN+30 /* filesize modified_time(24) filename(4)\n\0 */

extern int errno;

typedef struct _tind_var tind_t;
struct _tind_var {
  char tind_cwd[TIN_FLEN];    /* The directory the server started in */
};

void
tind_usage(char *progname)
{
  fprintf(stderr, 
          "Usage: %s [-p portnum] [-s password_directory]\n", 
          progname);
  exit(errno);
}

void
tind_logtime() 
{
  time_t tval;
  struct tm *cur_time;

  time(&tval);
  cur_time = localtime(&tval);
  fprintf(stderr, "%d:%d::%d:%d:%d ", cur_time->tm_mon, cur_time->tm_mday, 
      cur_time->tm_hour, cur_time->tm_min, cur_time->tm_sec);

  return;
}

/**
 * Looks up the user's password. Caller must deallocate
 * the returned buffer. On success, the password stored
 * in the appropriate file is returned; on failure,
 * NULL is returned and errno is set.
 *
 */
char * tind_find_pass(char * passdir, char * cwd, char * uniqname) {
  char * filename;
  char * password;
  FILE * passfile;
  struct stat fst;
  
  filename = (char *)malloc((strlen(cwd) + strlen(passdir) + 
                            strlen(uniqname) + 3));
  
  sprintf(filename, "%s/%s/%s", cwd, passdir, uniqname);
  
  if (stat(filename,&fst)) {
    free(filename);
    return NULL;
  }
  
  passfile = fopen(filename,"r");
  if (passfile == NULL) {
    free(filename);
    return NULL;
  }
  
  //allocate a buffer for the password, read it in
  password = (char *)malloc((fst.st_size+1)*sizeof(char));  
  
  //we ask for fst.st_size -1 bytes b/c the filesize
  //returned by stat() includes the eof.
  fread(password,1,(fst.st_size-1),passfile);
  password[fst.st_size-1] = '\0';
  
  //clean up
  fclose(passfile);
  free(filename);
  return password;
}

/*
 * tind_args: Parse args. (command line)
 *
 * Return 0 on success or errno on failure.
 * On successful return, "port" contains the
 * port# to use, in host byte order; "passdir"
 * contains the root directory; within this
 * directory are password files. Each password
 * file contains simply the user's password, and
 * is named with their uniqname.
 * 
*/
int
tind_args(int argc, char *argv[], int *port, char ** passdir)
{
  extern char *optarg;
  extern int optind;
  char c;

  errno = 0;

  *port = TIN_PORT;
  *passdir = TIND_PASSDIR;

  while ((c = getopt(argc, argv, "p:s:")) != EOF) {
    switch (c) {
    case 'p':
      *port = atoi(optarg);
      break;
    case 's':
      *passdir = optarg;
    default:
      errno = EINVAL;
      break;
    }
  }

  return(errno);
}

/*
 * tind_setup_conn: setup a TCP connection
 * bind it to port "port", listen on it.
 *
 * Terminate process on error.
 * Else return the connected socket id.
*/
int
tind_setup_conn(int port)
{
  int sd, reuse, err;
  struct sockaddr_in self;
  
  /* create a TCP socket */
  sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  tin_check_err(sd < 0, "socket");

  /* initialize socket address */
  memset((char *) &self, 0, sizeof(struct sockaddr_in));
  self.sin_family = AF_INET;
  self.sin_addr.s_addr = INADDR_ANY;
  self.sin_port = htons((unsigned short) port);

  /* reuse local address so that bind doesn't complain
     of address already in use. */
  reuse = 1;
  err = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse, sizeof(int));
  tin_check_err(err, "setsockopt");
  
  /* bind address to socket */
  err = bind(sd, (struct sockaddr *) &self, sizeof(struct sockaddr_in));
  tin_check_err(err, "bind");
  err = setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, (char *) &reuse, sizeof(int));
  tin_check_err(err, "setsockopt ALIVE");

  /* listen on socket */
  err = listen(sd, TIND_QLEN);
  tin_check_err(err, "listen");

  /* return socket id. */
  return(sd);
}

/*
 * tind_waitfor_conn: wait for connection on
 * the given socket, sd.
 *
 * Terminate process on error.
 * Else return the id of newly connected socket.
 * The socket is set to be non-blocking.
 * Upon return, the arg "chost" contains the host
 * entry of the connected client host.
*/ 
int
tind_waitfor_conn(int sd, struct hostent **chost)
{
  int td;
  int len;
  int flags;
  struct sockaddr_in client;
  
  /* wait for and accept the new connection. */
  len = sizeof(struct sockaddr_in);
  td = accept(sd, (struct sockaddr *) &client, (socklen_t*) &len);
  tin_check_err(td < 0, "accept");
  /* set socket non-blocking.  To prevent buggy client
     from h anging us.  See code and comments in _tind_recv()
     on EWOULDBLOCK. */

  flags = fcntl(td, F_GETFL, 0);
  tin_check_err(flags < 0, "fcntl getfl");
  flags |= O_NDELAY;
  flags = fcntl(td, F_SETFL, flags);
  tin_check_err(flags < 0, "fcntl setfl");

  /* get the host entry info on the connected host. */
  *chost = gethostbyaddr((char *) &client.sin_addr, sizeof(struct in_addr), AF_INET);
  tin_check_err(!*chost, "gethostbyaddr");
  /* for debug purposes */
  tind_logtime();
  fprintf(stderr, "Connected from %s\n", *chost ? (*chost)->h_name : 
"unknown client");

  /* return newly connected socket id. */
  return(td);
}

/*
 * tind_errmsg: allocate buffer of size TIN_MLEN
 * and put "msg" and "arg" in buffer.
 *
 * Return buffer.
 *
 * Caller must free returned buffer
 * and the arg "msg".
*/ 
char *
tind_errmsg(char *msg, unsigned long arg)
{
  char *errmsg;
  
  errmsg = (char *) malloc(TIN_MLEN);
  tin_check_err(!errmsg, "malloc");
  sprintf(errmsg, msg, arg);
  return(errmsg);
}

/*
 * tind_check_name: check that "name" doesn't
 * contain illegal characters (e.g. '/'),
 * and is not longer than "len."
 * If "lower" is 1, make "name" lower case.
 *
 * Return 1 if illegal character found.
 * Else return 0.
*/ 
int
tind_check_name(char *name, int len, int lower)
{
  int i;

  for (i = 0; i < len && name[i] != '\0'; i++) {
    if (name[i] == '/') {
      return(1);
    } else if (lower) {
      name[i] = tolower(name[i]);
    }
  }

  /* check for terminating null at <= len. */
  if (name[i] != '\0') {
    return(1);
  }

  return(0);
}

/*
 * tind_process_hdr: process the turnin header pointed to by tp.
 *
 * On client's error, return an error message that must be
 * freed by caller.  On server's system error, terminate
 * process.  Else return a NULL.
 *
 * Upon sucess, numbers in the turnin header are in host byte-order.
 * The directory "cwd/asgn/uniq" has been created and the
 * process is put in this directory.  The "cwd" directory
 * is the directory the process was started in.
*/ 
char *
tind_process_hdr(tih_t *tp, char *cwd, char * uniq)
{
  char *path;
  char *errmsg;

  /* check that version number is what we are expecting */
  if ((tp->tih_vers = ntohs(tp->tih_vers)) != TIN_VERS) {
    return(tind_errmsg("Version number %d not supported.\n", tp->tih_vers));
  }
           
  /* check that operation is known */
  tp->tih_op = ntohs(tp->tih_op);
  if (tp->tih_op != TIN_PUT && tp->tih_op != TIN_LST && tp->tih_op != TIN_SETPASS) {
    return(tind_errmsg("Unknown operation %d.\n", tp->tih_op));
  }
  
  /* check that asgn name doesn't contain any illegal character. */
  if (tind_check_name(tp->tih_asgn, TIN_ALEN-1, 1) || 
                      strlen(tp->tih_asgn) < TIN_ALEN-1 ||
                      strlen(tp->tih_asgn) >= TIN_ALEN) {
    return(tind_errmsg("Assignment name \"%s\" contains illegal character(s).\n",
                       (unsigned long) tp->tih_asgn));
  }
      
  if (tp->tih_op == TIN_PUT || tp->tih_op == TIN_SETPASS) {
    /* put nfiles in host byte-order. */
    tp->tih_nfiles = ntohs((unsigned short) tp->tih_nfiles);
    if ((tp->tih_nfiles < 1) || (tp->tih_nfiles > MAXSHORT)) {
      return(tind_errmsg("Number of files \"%d\" out of range.\n", (unsigned short) tp->tih_nfiles));
    }
  }

  /* check that cwd/asgn exists. */
  if (tp->tih_op != TIN_SETPASS) {
    errmsg = 0;
    path = (char *) malloc(strlen(cwd)+TIN_ALEN+strlen(uniq)+3);
    tin_check_err(!path, "malloc");
    sprintf(path, "%s/%s/", cwd, tp->tih_asgn);
  
    if (chdir(path)) {
      errmsg = tind_errmsg("Directory for %s doesn't exist.\nPlease notify instructor.\n",
                           (unsigned long) tp->tih_asgn);
  
    } else {
      /* if cwd/asgn exists, create cwd/asgn/uniq if it has not
         been previously created.  In either case, change dir to
         cwd/asgn/uniq. */
      strcat(path, uniq);
      if (chdir(path)) {
        if (tp->tih_op == TIN_LST) {
          errmsg = tind_errmsg("Directory for %s doesn't exist.\nHave you turned in any files?\n",
                               (unsigned long) uniq);
        } else if (mkdir(path, S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH) || chdir(path)) {
          errmsg = tind_errmsg("Directory for %s cannot be created.\nPlease notify instructor.\n",
                               (unsigned long) uniq);
        }
      }  
    }
  
    free(path);
  }
  return(errmsg);
}

/*
 * tind_process_fhdr: process file header pointed to by tp.
 *
 * On client's error, return an error message which the caller
 * must free.  On server's system error, terminate process.
 * Else return NULL.
 *
 * Upon sucess, the numbers in file header are
 * in host byte-order.
*/
char *
tind_process_fhdr(tif_t *tp)
{
  /* check that file name doesn't contain any illegal character. */
  if (tind_check_name(tp->tif_fname, TIN_FLEN-1, 0)) {
    return(tind_errmsg("File name \"%s\" contains illegal character.\n",
                       (unsigned long) tp->tif_fname));
  }

  /* put fsize in host byte-order. */
  tp->tif_fsize = ntohl((unsigned long) tp->tif_fsize);
  if ((tp->tif_fsize < 0) || (tp->tif_fsize > MAXLONG)) {
    return(tind_errmsg("File size \"%d\" out of range.\n", (unsigned long) tp->tif_fsize));
  }

  return(0);
}


/*
 * tind_recv: receive a buffer-full of data.
 *
 * Arg "buf" is where received data shall be
 * stored.  "Buf" is of "blen" size.
 *
 * In case of client side's error, an error message is
 * returned, which must be freed by the caller.
 * On server side's system error, terminate process.
 *
 * On success, a NULL is returned and
 * "blen" contains the actual length of data received.
*/ 
char *
tind_recv(int td, char *buf, int *blen)
{  
  extern char * _tind_recv(int, char *, int *);  

//  return(_tind_recv(td, buf, blen));
  

  fd_set rfds, efds; 
  int try=0;
  int received=0;
  int total_received=0;
  struct timeval to;
  int rval;

  while (try < TIND_MAXTRIES)
  {
  FD_ZERO (&rfds); //init
  FD_ZERO (&efds); //init  

  FD_SET(td,&rfds); //add td to listening set
  FD_SET(td,&efds); //add td to error set

  to.tv_sec =  TIND_SLEEPUSEC / 1000000;
  to.tv_usec = TIND_SLEEPUSEC % 1000000;
  
  // blocking call to select
  rval=select(td+1,&rfds,NULL,&efds,&to);

  if (rval<0)
  	{
	perror ("select");
	exit(1);
	}

   if (FD_ISSET (td, &efds)) //some error with client connection
   	{
	 //client closes connection too early
	 close (td); 
	 return tind_errmsg ("Connection closed\n",0); 
	}	

  if (FD_ISSET (td, &rfds)) //check if it is there
  	{
	received=recv(td, buf + total_received, *blen - total_received,0);

		if (received==-1)
			{
			 //client closes connection too early
			 close (td); 
			 return tind_errmsg ("Connection closed\n",0); 
			}
			
	   	else if (received != 0)
			{
			total_received += received;
			if (total_received == * blen) 
				{return NULL;} // correct termination 
			}
			
		else	{
			try++; //increment try only if nothing is received
			}
	}
   else
   	{
	  try++; //increment try only if nothing is received
	}
   }
 
 // if you are here then a client error had occured
 *blen=total_received;
 close(td);
 return tind_errmsg("Connection time-out",0);

 
}

/*
 * tind_recv_file: receive a file on socket td.
 * The file is described in the file header pointed to by tp.
 * Assume file header has been processed by tind_process_fhdr().
 *
 * Upon client side's error, return an error message that
 * the caller must free.  Upon server side system error,
 * terminate process.  Else return NULL.
 *
 * Upon sucess, file named in the file header is created
 * and its contents copied from client.  If a file of
 * the same name exists, clobber it.
*/ 
char *
tind_recv_file(int td, tif_t *tp)
{
  int fd;
  char *errmsg;
  int bread, toread;
  char buf[TIN_DLEN];

  /* open file, replacing current content. */
  fd = open(tp->tif_fname, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP);
  if (fd < 0) {
    return(tind_errmsg("File \"%s\" cannot be created.\nPlease notify instructor.",
                       (unsigned long) tp->tif_fname));
  }

  /* receive file, one TIN_DLEN block at a time. */
  errmsg = 0;
  toread = tp->tif_fsize;
  while (toread > 0 && !errmsg) {
    /* the last block may be smaller than TIN_DLEN. */
    bread = toread < TIN_DLEN ? toread : TIN_DLEN;
    errmsg = tind_recv(td, buf, &bread);
    if (!errmsg) {
      write(fd, buf, bread);
      toread -= bread;
    }
  }

  close(fd); 
  return(errmsg);
}

/*
 * tind_send_list: send back a list of files in
 * the assignment directory to client.
*/
void
tind_send_list(int td, char *progname, tih_t *tp, char * uniq)
{
  char * hname;
  char flist[TIN_DLEN];
  char *flp;
  DIR *dp;
  struct dirent *dirp;
  struct stat fst;
  int tosend, sent;
  
  /* identify self */
  hname = getenv("HOSTNAME");
  
  flp = flist;
  if (hname != NULL) flp += sprintf(flp, "%s@%s: %s/%s\n", progname, hname, tp->tih_asgn, uniq);
  else flp += sprintf(flp, "%s@??: %s/%s\n", progname, tp->tih_asgn, uniq);
  
  tin_check_err(!(dp = opendir(".")), "opendir");

  for (;;) {
    dirp = readdir(dp);
    if (!dirp || ((flp + TIND_FLIST) > &flist[TIN_DLEN])) {
      /* Send out what has been accumulated.
         Since we put the socket in
         async mode, make sure all is sent. */
      if (!dirp) *flp++ = '\0';
      sent = 0;
      tosend = flp - flist;
      flp = flist;
      /*add to log*/
      fprintf(stderr, "%s\n", flp );
      while (tosend && !(sent < 0)) {
        sent = send(td, flp, tosend, 0);
        /* don't die just because client is dead */
        tin_check_err(sent < 0 && errno != EPIPE, "send");
        flp += sent;
        tosend -= sent;
      }
      if (!dirp) {
        break;
      } else {
        flp = flist;
      }
    }
    stat(dirp->d_name, &fst);
    if (S_ISREG(fst.st_mode)) {
      flp += sprintf(flp, "%d\t", (int)fst.st_size);
      memcpy(flp,ctime(&fst.st_mtime),24);
      flp += 24;
      flp += sprintf(flp, "\t%s\n", dirp->d_name);
    }
  }
  return;
}

/*
 * Reads user's new password off of td, and stores it
 * in their password file.
 *
 * Returns result of operation (as message).
 */
char * tind_recv_pass(int td, int blen, char * passdir, char * cwd, char * uniq) {
  char * oldpass;
  char * newpass;
  char * errmsg;
  char * filename;
  FILE * passfile;
fprintf(stderr,"newpass len=%d\n",blen);
  if (blen < TIN_FLEN && blen >= 4) {
    newpass = (char *)malloc(blen + 1);
    oldpass = tind_find_pass(passdir, cwd, uniq);
    if (!oldpass) {
      return(tind_errmsg("Can't find old password for user %s\n",(unsigned long)uniq));
    }
    errmsg = tind_recv(td,newpass,&blen);
    if (!errmsg) {
      //got new password. open the file and write it
      newpass[blen] = '\0';
      filename = (char *)malloc((strlen(cwd) + strlen(passdir) + 
                            strlen(uniq) + 3));
  
      sprintf(filename, "%s/%s/%s", cwd, passdir, uniq);
fprintf(stderr,"decrypting password\n");
      encrypt_pass(newpass,oldpass,blen);
fprintf(stderr,"newpass=%s\n",newpass);

      passfile = fopen(filename,"w");
      if (passfile == NULL) {
        free(oldpass);
        free(filename);
        free(newpass);
        return tind_errmsg("Could not open password file %s\n",(unsigned long) filename);;
      }

      free(filename);
      fprintf(passfile,"%s\n",newpass);
      fclose(passfile);
    } else {
      free(oldpass);
      free(newpass);
      return tind_errmsg("Did not receive %d bytes as expected\n",blen);
    }
    free(oldpass);
    free(newpass);
    errmsg = tind_errmsg("New password successfully written for user %s\n",(unsigned long) uniq);
  } else {
    if (blen < 4) errmsg = tind_errmsg("Password length %d too short\n",blen);
    else errmsg = tind_errmsg("Password length %d too long\n",blen);
  }
  return errmsg;
}

/*
 * tind_handshake: Performs the server side of the 
 * authentication handshake.
 *
 * On success, the user is authenticated and
 * their uniqname is put in the buffer pointed to
 * by uniq. NULL is returned. The caller must free
 * the uniq buffer.
 *
 * On failure, an error message is returned which the
 * caller must free. The caller should terminate
 * the connection (i.e., close the socket).
 */
char * tind_handshake(int td, char * passfn, char * cwd, char ** uniq) {
  char buf[TIND_CHALLENGE_LEN];
  char nonce[3];
  char * pass;
  char * errmsg = (char *)0;
  char * token;
  int sent, tosend = 0, rval;
  int blen = TIN_CHALLENGE_LEN;
  memset(nonce,0,3);
  
  errmsg = tind_recv(td,buf,&blen);

  if (!errmsg) {
    //break out the uniqname
    *uniq = (char *)malloc(9);
    (*uniq)[8] = '\0';
    strncpy(*uniq,buf,8);

    //find the user's password
    pass = tind_find_pass(passfn, cwd, *uniq);
    if (pass == NULL) {
      free(*uniq);
      return(tind_errmsg("Password for user \"%s\" could not be found. Check your password and try again.",
                       (unsigned long) *uniq));
    }

    //Compute our token
    strncpy(nonce,&(buf[9]),2);
    nonce[2] = '\0'; //crypt expects a string
    token = crypt(pass,nonce);

    //create our nonce to authenticate the client
    rval = create_nonce(nonce);
    if (rval) {
      free(*uniq);
      free(pass);
      return(tind_errmsg("Could not create server challenge nonce, code %d.",rval));
    }
    
    //copy the token data into the buffer (which we're done with)
    strncpy(buf,token,CRYPT_LEN);
    strncpy(buf+CRYPT_LEN,nonce,2);

    //send our response (the second step of authentication)
    tosend = TIND_CHALLENGE_LEN;
    sent = 0;
    while (tosend && !(sent < 0)) {
      sent = send(td, &(buf[sent]), tosend, 0);
      tin_check_err(sent < 0 && errno != EPIPE, "send");
      tosend -= sent;
    }
    if (sent < 0) {
      free(*uniq);
      return(tind_errmsg("Could not send challenge, errno=%d.",errno));
    }
    
    //create the correct client response
    //for comparison
    token = crypt(pass,nonce);
    //read the client response
    blen = CRYPT_LEN;

    errmsg = tind_recv(td,buf,&blen);
    if (errmsg) {
      free(pass);
      free(*uniq);
      return errmsg;
    }

    buf[CRYPT_LEN] = '\0'; //null-terminate the crypt
    if (strncmp(token,buf,CRYPT_LEN)) {
      //client response did not match
      free(pass);
      free(*uniq);
      return(tind_errmsg("Client response was %s, it did not match.",(unsigned long)buf));
    }
    
    free(pass);
  }
  return errmsg;
}

int
main(int argc, char *argv[])
{
  int port;
  tih_t tih;
  tif_t tif;
  tind_t tind;
  int sd, td;
  int i, len;
  char *errmsg, *errp;
  struct hostent *chost;
  int tosend, sent;
  char *passdir, *uniq;

  if (tind_args(argc, argv, &port, &passdir)) {
    tind_usage(argv[0]);
  }

  /* get the working directory the process was started in. */
  tin_check_err(!getcwd(tind.tind_cwd, TIN_FLEN), "getcwd");

  signal(SIGPIPE, SIG_IGN);        /* don't die just because client is dead */

  umask(0);

  sd = tind_setup_conn(port);
  fprintf(stderr, "Waiting for client to connect . . . .\n"); 
  
  while (1) {

    td = tind_waitfor_conn(sd, &chost);

    
    errmsg = tind_handshake(td, passdir, tind.tind_cwd, &uniq);
    if (!errmsg) {
      tind_logtime();
      fprintf(stderr, "Authenticated %s\n",uniq);
      /* Authentication successful, proceed */
      /* grab the client's turnin header. */
      len = sizeof(tih_t);
      errmsg = tind_recv(td, (char *) &tih, &len);

      if (!errmsg) {
        errmsg = tind_process_hdr(&tih, tind.tind_cwd, uniq);

        /* add to log */
        tind_logtime();
        if (tih.tih_op == TIN_PUT) {
          /* for each file, up to nfiles, grab the file. */
          fprintf(stderr, "PUT %d file(s).\n", tih.tih_nfiles);
          for (i = 1; !errmsg && i <= tih.tih_nfiles; i++) {
            len = sizeof(tif_t);
            errmsg = tind_recv(td, (char *) &tif, &len);
            if (!errmsg) {
              errmsg = tind_process_fhdr(&tif);
              if (!errmsg) {
                errmsg = tind_recv_file(td, &tif);
              }
            }
          }
        } else if (tih.tih_op == TIN_SETPASS) {
          errmsg = tind_recv_pass(td, tih.tih_nfiles, passdir, tind.tind_cwd, uniq);
         } else {
          fprintf(stderr, "LIST.\n");
        }
      }

      if (!errmsg) {
        /* if all went well, tell the client how many
           files have been received for which assignment. */
        tind_send_list(td, argv[0], &tih, uniq);
      } else {
        /* send errmsg.  Since we put the socket in
           async mode, make sure all is sent. */
        fprintf(stderr, "%s", errmsg);
        errp = errmsg;
        tosend = strlen(errmsg)+1;
        sent = 0;
        while (tosend && !(sent < 0)) {
          sent = send(td, errp, tosend, 0);
          /* don't die just because client is dead */
          tin_check_err(sent < 0 && errno != EPIPE, "send");
          errp += sent;
          tosend -= sent;
        }
        
        free(errmsg);
      free(uniq);
      }
    }
    
    sleep(TIND_ERRSLEEPSEC); /* wait for client to get msg */
    close(td);
  }

  exit(0);
}
